/*
 * General purpose scatter-gather DIO API
 * Please refer to include/linux/sgat-dio.h for the detailed Documentation
 * 
 */
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/scatterlist.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif /* LINUX_VERSION_CODE */
#include <linux/iobuf.h>
#include <linux/sgat-dio.h>

#define sgat_dio_cptouser sgat_dio_cpfrmuser

static void *sgat_dio_malloc(int rqSz);
static void sgat_dio_free(void * buff);
static void expand_sclist(struct sgat_dio_fops *fops,sgat_dio_req *dio_req, int sc_elems);
static void *add_dio_chunk(struct sgat_dio_fops *fops, sgat_dio_req * dio_req, 
                    u_long address, int length,void *sclp);
static int end_dio_req(struct sgat_dio_fops *fops, sgat_dio_req *dio_req);

static void sgat_dio_done(struct sgat_dio_fops *fops, sgat_dio_req *dio_req );
static void sgat_dio_done_nocopy(struct sgat_dio_fops *fops, sgat_dio_req *dio_req);
static int sgat_dio_cpfrmuser(u_char *src, u_char *dest, int size);

static int sgat_dio_cpfrmuser(u_char *src, u_char *dest, int size){
    int i;
    u_char *s = src;
    u_char *d = dest;

    if ( !s || !d )
        return -EINVAL;

    for ( i =0; i < size; i++){
        if (__get_user_nocheck(*d, s, sizeof(u_char))){
            DIO_DPRINTK("failed verify_area\n");
            return -EINVAL;
        }
        s++;
        d++;
    }
    return 0;
}
static void *sgat_dio_malloc(int rqSz)
{
    char *resp = NULL;

    if (rqSz <= 0)
        return resp;
    resp = kmalloc(rqSz, GFP_KERNEL);
    if (resp) 
        memset(resp, 0, rqSz);
    DIO_DPRINTK("size=%d,ret=0x%p\n", rqSz, resp);
    return resp;
}

static void sgat_dio_free(void * buff)
{
    if (!buff) 
        return;
    DIO_DPRINTK("Free'ing %p\n",buff);
    kfree(buff);  
}

static void expand_sclist(struct sgat_dio_fops *fops,sgat_dio_req *dio_req, int sc_elems)
{
    int elem_sz = sizeof(struct scatterlist);
    if ( dio_req ){
        if ( !dio_req->sclist || (sc_elems > dio_req->sc_elems) ){
            if ( dio_req->sclist) 
                fops->sgat_dio_free((void *)dio_req->sclist);
            dio_req->sclist = fops->sgat_dio_malloc(sc_elems * elem_sz);
            dio_req->sc_elems = sc_elems;
        }
    }
}

static void *add_dio_chunk(struct sgat_dio_fops *fops, sgat_dio_req * dio_req, 
                    u_long address, int length,void *sclp){
    (void) fops;
    struct  scatterlist *p = (struct  scatterlist *) sclp;
    dio_req->k_useg++;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,13)
    p->page = NULL;
#endif
    p->address = (char *)address;
    p->length = length;
    p++;
    return p;
}

static int end_dio_req(struct sgat_dio_fops *fops, sgat_dio_req *dio_req) {
    (void) fops;
    dio_req->in_use = 1;
    return 0;
}

static void sgat_dio_done(struct sgat_dio_fops *fops, sgat_dio_req *dio_req )
{
    if (!dio_req )
        return;	
    if ( dio_req->k_io_hdrp &&  dio_req->u_io_hdrp ){
        sgat_dio_cptouser((u_char *)dio_req->k_io_hdrp,
                        (u_char *)dio_req->u_io_hdrp,sizeof(sgat_dio_hdr_t));
    }
    dio_req->in_use = 0;
}

static void sgat_dio_done_nocopy(struct sgat_dio_fops *fops, sgat_dio_req *dio_req)
{
    if (!dio_req )
        return;	
    dio_req->in_use = 0;
    return;	
}

int sgat_dio_get_default_fops(struct sgat_dio_fops *fops)
{
    if ( !fops){
        DIO_DPRINTK("Invalid parameter\n");
        return -EINVAL;
    }
                    
    fops->sgat_dio_malloc = sgat_dio_malloc;
    fops->sgat_dio_free = sgat_dio_free;
    fops->add_dio_chunk = add_dio_chunk;
    fops->end_dio_req = end_dio_req;
    fops->sgat_dio_done = sgat_dio_done;
    fops->expand_sclist = expand_sclist;

    return 0;
}

static int compute_scbuf_size(struct iovec *iovecp,int iovec_count)
{
    struct iovec *p;
    int	i;
    int scl_entries,sg_tablesize = 0;

    if (!iovecp)
        return -EINVAL;	
    for (i= 0,p = iovecp;i < iovec_count ; i++,p++) {
        scl_entries = (p->iov_len / PAGE_SIZE) * 2 + 1;
        if ( 1 == scl_entries ) {
            scl_entries = 2 ;
        }
        sg_tablesize += scl_entries;
    }
    return sg_tablesize;
}

static int get_dxfer_len(struct iovec *iovecp,int iovec_count){
    struct iovec *p;
    int	i;
    int     total_len = 0;

    if (!iovecp)
        return -EINVAL;	

    for (i= 0,p = iovecp;i < iovec_count ; i++,p++){
        total_len += p->iov_len;
    }
    return total_len;
}

sgat_dio_req *sgat_dio_alloc_req(struct sgat_dio_fops *fops)
{
    sgat_dio_req *dio_req;

    if ( !fops )
        return NULL;

    dio_req = (sgat_dio_req *) fops->sgat_dio_malloc(sizeof(sgat_dio_req));    
    if ( !dio_req )
        return NULL;
    return dio_req;
}

int sgat_dio_init_req(struct sgat_dio_fops *fops,
                        sgat_dio_req *dio_req,
                        sgat_dio_hdr_t *u_io_hdrp)
{
    int res;

    if ( !fops || !u_io_hdrp || !dio_req)
        return -EINVAL;

    dio_req->sc_elems = 0;
    dio_req->u_io_hdrp = NULL;
    dio_req->k_io_hdrp = NULL;
    dio_req->kiobuf_inuse = 0;
    dio_req->k_useg = 0;

    dio_req->k_io_hdrp = (sgat_dio_hdr_t *) fops->sgat_dio_malloc(sizeof(sgat_dio_hdr_t));
    if ( !dio_req->k_io_hdrp ){
        DIO_DPRINTK("sgat_dio_malloc failed\n");
        goto errout;
    }

    if ( sgat_dio_cpfrmuser((u_char *)u_io_hdrp, 
            (u_char *)dio_req->k_io_hdrp,sizeof(sgat_dio_hdr_t))){
        goto errout;
    }

    dio_req->u_io_hdrp = u_io_hdrp;

    res = alloc_unbacked_kiovec(1, &dio_req->kiobp);
    if (0 != res) {
        DIO_DPRINTK("alloc_kiovec res=%d\n", res);
        goto errout;
    }

    dio_req->sc_elems = compute_scbuf_size(dio_req->k_io_hdrp->iovec,
                                            dio_req->k_io_hdrp->iovec_count);
    DIO_DPRINTK("sc_elems %d\n",dio_req->sc_elems);
    if (dio_req->sc_elems <= 0 ) {
        DIO_DPRINTK("Invalid sc_elems %d\n",dio_req->sc_elems);
        goto errout;
    }
    dio_req->sclist = fops->sgat_dio_malloc(dio_req->sc_elems * sizeof(struct scatterlist));
    if ( !dio_req->sclist )
        goto errout;

    return 0;
errout:
    sgat_dio_exit_req(fops, dio_req,0);
    return -ENOMEM;
}

void sgat_dio_free_req(struct sgat_dio_fops *fops, struct sgat_dio_req *dio_req)
{
    if ( dio_req)
        fops->sgat_dio_free((void *)dio_req);
}

void sgat_dio_exit_req(struct sgat_dio_fops *fops, struct sgat_dio_req *dio_req, int flushDcache)
{
    if ( dio_req) {
        if (dio_req->kiobuf_inuse && dio_req->kiobp) {
            unlock_pages(dio_req->kiobp->maplist);
            if ( flushDcache){
                flush_dcache();
            }

            dio_req->kiobp->nr_pages = 0;
            dio_req->kiobp->length  = 0;
            dio_req->kiobuf_inuse = 0;
            DIO_DPRINTK("unmapping user pages\n");
        }

        if ( dio_req->kiobp ){
            free_kiovec(1, &dio_req->kiobp);
            dio_req->kiobp = NULL;
        }

        if ( dio_req->sclist ){
            fops->sgat_dio_free((void *)dio_req->sclist);
            dio_req->sclist = NULL;
        }

        if ( dio_req->k_io_hdrp ){
            fops->sgat_dio_free((void *)dio_req->k_io_hdrp);
            dio_req->k_io_hdrp = NULL;
        }
    }
}

static int sgat_dio_build_sclist(struct sgat_dio_fops *fops,
                            sgat_dio_req *dio_req)
{
    int     k,l;
    int     offset, rem_sz,useg_len,toffset;
    struct  kiobuf *kp;
    void    *sclp;
    int	    pgindex =0;
    int     pgcount=0;
    unsigned short iovec_count=0;
    int     res;
    int     dxfer_len;
    unsigned long p;
    int     err;
    unsigned long kaddress;
    int     length;
    struct  iovec *iovecp ;
    unsigned short iovecc;
    int     ulen;
    struct page *start_pg, *next_pg, *cur_pg;
    
    if (!fops || !dio_req || 
         !dio_req->kiobp || !dio_req->k_io_hdrp ||
         !dio_req->k_io_hdrp->iovec ){
            return -EFAULT;
    }

    kp = dio_req->kiobp; 

    iovecp = dio_req->k_io_hdrp->iovec;
    iovecc = dio_req->k_io_hdrp->iovec_count;
                    
    fops->expand_sclist(fops,dio_req, iovecc); 
    sclp = dio_req->sclist;
    if (!sclp)
        return -ENOMEM;

    dxfer_len = get_dxfer_len(iovecp, iovecc);

    for (k = 0, rem_sz = dxfer_len,iovec_count = 0,pgindex = 0;
        (iovec_count < iovecc) && (rem_sz > 0 ) && (k < iovecc);
        iovec_count++, k++) {

        p = (unsigned long) iovecp[iovec_count].iov_base;
        if ( !p )
            return -EFAULT;
        
        useg_len = iovecp[iovec_count].iov_len;
        offset = p & (PAGE_SIZE-1);
        pgcount = (p + useg_len + PAGE_SIZE - 1)/PAGE_SIZE - p/PAGE_SIZE;
        if (!pgcount) 
            BUG();

        if ((err = verify_area( (dio_req->k_io_hdrp->dxfer_dir == READ) ? 
                            VERIFY_WRITE:VERIFY_READ, 
                            (u_char *)p, iovecp[iovec_count].iov_len))){
            return err;
        }

        DIO_DPRINTK ("iovec %d offset %d len"
                    "%d pgcount %d pgindex %d rem_sz %d\n",
                    iovec_count,offset,useg_len,pgcount,pgindex,rem_sz);

        for( length = 0,ulen = useg_len,l = pgindex, toffset = offset,
            next_pg = NULL, start_pg = kp->maplist[l];
            (ulen > 0) && (start_pg != NULL) &&
            (l < (pgindex+pgcount)); 
            l++ ) {

            kaddress = (unsigned long) page_address(start_pg)+offset;

            cur_pg = kp->maplist[l];
            next_pg = ((l+1) < (pgindex+pgcount))? 
                            kp->maplist[l+1]:NULL;
                
            if ( next_pg || (ulen / PAGE_SIZE)){
                ulen -= PAGE_SIZE - toffset;
                length += PAGE_SIZE - toffset;
                toffset = 0;
                if ( (cur_pg + PAGE_SIZE) == next_pg)
                    continue;
            }
            else {
                /* Last page */
                length += ulen;
                ulen = 0;
            }
            sclp = fops->add_dio_chunk(fops,
                                    dio_req,
                                    kaddress, 
                                    length,sclp);
            if ( !sclp ) 
                return -EFAULT;
        
            DIO_DPRINTK("k=%d, a=0x%p, len=%d"
                        "ulen = %d\n", k,
                        kaddress, length, ulen);
            length = 0;
            offset = 0;
            toffset = 0;
            start_pg = next_pg;
        }
        pgindex += pgcount;
        rem_sz -= useg_len;
        DIO_DPRINTK("rem_sz=%d\n",rem_sz,k);
    }
    if (rem_sz > 0)
        return -EFAULT; 

    res = fops->end_dio_req(fops,dio_req);
    if (res < 0 )
        return -EFAULT;
        
    DIO_DPRINTK("ku_seg=%d, rem_sz=%d\n",k, rem_sz);
    return 0;
}

int sgat_dio_build_sg_list(struct sgat_dio_fops *fops,
                           sgat_dio_req *dio_req, int flushDcache)
{

    int res;
    sgat_dio_hdr_t  *k_io_hdrp= dio_req->k_io_hdrp;

    if ( !dio_req || !k_io_hdrp)
        return -EINVAL;

    if ( !k_io_hdrp->iovec_count || !k_io_hdrp->iovec) {
        DIO_DPRINTK("invalid inputs(iovec count %d )\n",
                k_io_hdrp->iovec_count);
        sgat_dio_free_req(fops,dio_req);
        return -EINVAL;
    }

    res = sgat_dio_map_user_kiobuf(k_io_hdrp->dxfer_dir, dio_req->kiobp, 
                                    k_io_hdrp->iovec, k_io_hdrp->iovec_count,
                                    flushDcache);
    if (0 != res) {
        DIO_DPRINTK("map_user_kiobuf res=%d\n", res);
        sgat_dio_free_req(fops,dio_req);
        return -EFAULT;
    }

    dio_req->kiobuf_inuse = 1;
    res = sgat_dio_build_sclist(fops, dio_req);
    if (res < 0) {
        sgat_dio_free_req(fops,dio_req);
        return res;
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
/*
 * Fault a page in and (atomically) increment its reference count, to
 * guarantee that it stays physically present
 */ 
struct page *lock_page_flush_dcache_pg(struct mm_struct *mm, 
                            struct vm_area_struct *vma,
                            unsigned long addr, int write)
{
    struct page *page = lock_down_page(mm, vma, addr, write);
    flush_dcache_page(page);
    // XXX: also need to unguard when guard implemented
    return page;
}

struct page *lock_down_page(struct mm_struct *mm, struct vm_area_struct *vma,
                            unsigned long addr, int write)
{
    struct page *page;

    if (vma->vm_start > addr) {
        if (!(vma->vm_flags & VM_GROWSDOWN) || expand_stack(vma, addr)) {
            return NULL;
        }
    }

    /* The common case is that the page is already present (and dirty if
     writing to mem), only call handle_mm_fault if not so.*/
    while (1) {
        pgd_t *pgd;
        pmd_t *pmd;

        spin_lock(&mm->page_table_lock);
        pgd = pgd_offset(current->mm, addr);
        pmd = pmd_offset(pgd, addr);
        if (pmd) {
            pte_t * pte = pte_offset(pmd, addr);
            if (pte && pte_present(*pte) && (!write || pte_write(*pte))) {
                page = pte_page(*pte);
                break;
            }
        }
        spin_unlock(&mm->page_table_lock);
        if (handle_mm_fault(mm, vma, addr, write, current->rt_priority) <= 0) {
                return NULL;
        }
    }
    spin_lock(&mm->page_table_lock);
    get_page(page);
    spin_unlock(&mm->page_table_lock);
    // XXX: also need to unguard when guard implemented
    return page;
}

void unlock_pages(struct page **page_list)
{
    while (*page_list) {
        put_page(*page_list);
        page_list++;
    }
    // XXX: also need to unguard when guard implemented
}

int sgat_dio_map_user_kiobuf(int rw, struct kiobuf *iobuf, 
                             struct iovec *iovec, int iovec_count, 
                             int flushDcache)
{
    int pgcount=0;
	int total_pages=0;
    int total_len=0;
	int count,err;
    int bytecount; 
	struct mm_struct *mm;
	struct iovec *tiovec;	
    unsigned long curr_ptr;
    unsigned int curr_cnt;
    struct page **page_list;
    struct page *chunk_page;
    struct vm_area_struct *vma;
    unsigned long vma_end;
    int chunk_size;
    int x=0;

	mm = current->mm;
    
	iobuf->locked = 0;
    iobuf->nr_pages = 0;
    iobuf->length  = 0;

	for ( count = 0, tiovec = iovec; count < iovec_count; count++, tiovec++){
        pgcount = (((u_long)tiovec->iov_base) + tiovec->iov_len + PAGE_SIZE - 1)/PAGE_SIZE 
                    - ((u_long)tiovec->iov_base)/PAGE_SIZE;
        total_pages += pgcount;
        total_len += tiovec->iov_len;
    }

	err = expand_kiobuf(iobuf, total_pages+1);
	if (err)
		return err;
        
    iobuf->nr_pages = total_pages;
    iobuf->length = total_len;

	/* Try to fault in all of the necessary pages */
	down_read(&mm->mmap_sem);

    page_list = iobuf->maplist;
	for ( count = 0, tiovec = iovec; count < iovec_count; count++, tiovec++){

        curr_ptr = (unsigned long)tiovec->iov_base;
        curr_cnt = tiovec->iov_len;

        /* force a vma_find() first time through the loop below*/
        vma = NULL;
        vma_end = 0;

        bytecount = 0;
        while ((int)curr_cnt > bytecount) {

            if (curr_ptr >= vma_end) {
                vma = find_vma(mm, curr_ptr);
                if (!vma) {
                    err = -EFAULT;
                    goto err_out;
                }
                /* READ means write into memory , read from device*/
                if (!(vma->vm_flags & ((rw==READ) ? VM_WRITE : VM_READ))) {
                    err = -EACCES;
                    goto err_out;
                }
                vma_end = vma->vm_end;
            }

            /*Forcing physical pages allocation, COW*/
            if(rw==READ)
                put_user(x,(char *)curr_ptr);
            /* Make the page present, and bump up its ref-count.*/
            chunk_page = lock_down_page(mm, vma, curr_ptr, (rw==READ));
            if (!chunk_page) {
                err = -EFAULT;
                goto err_out;
            }

            *(page_list++) = chunk_page;

                /* Check to see if the current pointer is page-aligned...*/
            if (curr_ptr % PAGE_SIZE) {
                        /* ...it wasn't page aligned*/
                chunk_size = ((PAGE_SIZE - (curr_ptr % PAGE_SIZE)) < (curr_cnt - bytecount)) ?
                            (PAGE_SIZE - (curr_ptr % PAGE_SIZE)):(curr_cnt - bytecount);
            } else {
                        /* ...it is page aligned*/
                chunk_size = (PAGE_SIZE < (curr_cnt - bytecount)) ?
                            PAGE_SIZE:(curr_cnt - bytecount);
            }

            bytecount += chunk_size; 
            curr_ptr += chunk_size;
        }
    }
    *page_list = 0;

    if ( flushDcache ) 
        flush_dcache();

err_out:	
    if (err < 0) {
        unlock_pages(page_list);
        if (flushDcache ) {
            flush_dcache();
        }

        iobuf->nr_pages = 0;
        iobuf->length  = 0;
        up_read(&mm->mmap_sem);
        return err;
    }
	up_read(&mm->mmap_sem);

	return 0;
}
